# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Utility functions."""

from functools import wraps
from urllib.parse import urlparse

from flask import request
from onelogin.saml2.auth import OneLogin_Saml2_Auth

from invenio_saml.proxies import current_sso_saml


def prepare_flask_request(request):
    """Prepare OneLogin-friendly request."""
    # If server is behind proxys or balancers use the HTTP_X_FORWARDED fields
    uri_data = urlparse(request.url)
    return {
        "get_data": request.args.copy(),
        "http_host": request.host,
        "https": "on" if request.scheme == "https" else "off",
        "post_data": request.form.copy(),
        "script_name": request.path,
        "server_port": uri_data.port,
        # Uncomment if using ADFS as IdP,
        # https://github.com/onelogin/python-saml/pull/144
        # 'lowercase_urlencoding': True,
    }


def run_handler(handler_name):
    """Find a handler inside configuration and call it."""

    def decorated(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            res = f(self, *args, **kwargs)
            handler = current_sso_saml.get_handler(self.idp, handler_name)
            if handler:
                return handler(self, res)
            return res

        return inner

    return decorated


class SAMLAuth(OneLogin_Saml2_Auth):
    """Encapsulate OneLogin SP SAML instance."""

    def __init__(self, idp, settings, *args, **kwargs):
        """Initialization."""
        self.idp = idp
        self._settings = settings
        req = current_sso_saml.prepare_flask_request(request)
        super(SAMLAuth, self).__init__(req, self._settings, *args, **kwargs)

    @run_handler("settings_handler")
    def get_settings(self):
        """Get settings info and call handler.

        :return: ``OneLogin_Saml2_Setting`` object
        """
        settings = super(SAMLAuth, self).get_settings()
        return settings

    @run_handler("login_handler")
    def login(self, *args, **kwargs):
        """Wrapper around ``OneLogin_Saml2_Auth.login``."""
        next_url = super(SAMLAuth, self).login(*args, **kwargs)
        return next_url

    @run_handler("logout_handler")
    def logout(self, *args, **kwargs):
        """Wrapper around ``OneLogin_Saml2_Auth.logout``."""
        next_url = super(SAMLAuth, self).logout(*args, **kwargs)
        return next_url

    @run_handler("acs_handler")
    def acs_handler(self, next_url):
        """Call ACS handler from config."""
        return next_url

    @run_handler("sls_handler")
    def sls_handler(self, next_url):
        """Call SLS handler from config."""
        return next_url



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Graz University of Technology.
# Copyright (C) 2022-2024 Esteban J. G. Gabancho.
#
# invenio-saml is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""SAML integration functions."""

from functools import wraps

from flask import (
    Blueprint,
    abort,
    current_app,
    jsonify,
    make_response,
    redirect,
    request,
    session,
)

from invenio_saml.errors import IdentityProviderNotFound
from invenio_saml.proxies import current_sso_saml


def verify_idp(f):
    """Check if the Identity Provider is correctly exists."""

    @wraps(f)
    def inner(idp, *args, **kwargs):
        try:
            return f(idp=idp, auth=current_sso_saml.get_auth(idp), *args, **kwargs)
        except IdentityProviderNotFound:
            # IdP name not found inside the configuration
            return abort(404, "Identity Provider not found")

    return inner


@verify_idp
def metadata(idp, auth):
    """Expose XML configuration of the Service Provider (us)."""
    current_app.logger.debug("Handling metadata request for {}".format(idp))

    settings = auth.get_settings()
    sp_metadata = settings.get_sp_metadata()
    errors = settings.validate_metadata(sp_metadata)

    if errors:
        error_reason = auth.get_last_error_reason()
        if error_reason:
            errors.append(auth.get_last_error_reason())
            current_app.logger.error("Handling metadata request: {}".format(errors))
        return jsonify(errors), 401
    else:
        current_app.logger.debug("Metadata request response: {}".format(sp_metadata))
        resp = make_response(sp_metadata, 200)
        resp.headers["Content-Type"] = "text/xml"
        return resp


@verify_idp
def sso(idp, auth):
    """Send user to IdP login page (SAML single sign-on)."""
    current_app.logger.debug("SSO SAML for {}".format(idp))

    next_url = request.args.get("next", request.referrer) or request.host_url
    login = auth.login(return_to=next_url)

    current_app.logger.debug(
        "SSO Request XML: \n{}".format(auth.get_last_request_xml())
    )
    current_app.logger.debug('Redirecting to "{}" to initiate login'.format(login))

    if not login:
        # This should never happen, but just in case
        abort(401)

    return redirect(login)


@verify_idp
def acs(idp, auth):
    """Authorized handler callback (Assertion Consumer Service).

    It gets called by the IdP with SAML assertion when authentication has been
    performed.
    """
    try:
        # TODO https://github.com/onelogin/python3-saml/issues/39 ?
        auth.process_response()
    except Exception:  # TODO better exception handling
        return abort(400)

    errors = auth.get_errors()

    if errors:
        error_reason = auth.get_last_error_reason()
        if error_reason:
            errors.append(auth.get_last_error_reason())
            current_app.logger.error("Handling ACS request: {}".format(errors))
        return jsonify(errors), 401

    if not auth.is_authenticated():
        return abort(403)

    # Set SSO specific IdP metadata in the session, (used later in slo)
    session[current_app.config["SSO_SAML_SESSION_KEY_NAME_ID"]] = auth.get_nameid()
    session[current_app.config["SSO_SAML_SESSION_KEY_SESSION_INDEX"]] = (
        auth.get_session_index()
    )

    next_url = auth.acs_handler(request.form.get("RelayState")) or "/"

    return redirect(next_url)


@verify_idp
def slo(idp, auth):
    """Send user to IdP logout page (SAML single logout)."""
    name_id = session.get(current_app.config["SSO_SAML_SESSION_KEY_NAME_ID"])
    session_index = session.get(
        current_app.config["SSO_SAML_SESSION_KEY_SESSION_INDEX"]
    )

    next_url = request.args.get("next", request.referrer) or request.host_url

    logout = auth.logout(
        return_to=next_url, name_id=name_id, session_index=session_index
    )

    current_app.logger.debug(
        "SLO Request XML: \n{}".format(auth.get_last_request_xml())
    )
    current_app.logger.debug('Redirecting to "{}" to initiate logout'.format(logout))

    if not logout:
        # This should never happen, but just in case
        abort(401)

    return redirect(logout)


@verify_idp
def sls(idp, auth):
    """Logout handler callback (Single Logout Service).

    It Consumes LogoutResponse from IdP when logout has been performed.
    """
    # Process the SLO message received from IdP
    next_url = auth.process_slo(delete_session_cb=lambda: session.clear())

    errors = auth.get_errors()
    if errors:
        error_reason = auth.get_last_error_reason()
        if error_reason:
            errors.append(auth.get_last_error_reason())
            current_app.logger.error("Handling SLS request: {}".format(errors))
        return jsonify(errors), 401

    next_url = auth.sls_handler(next_url) or "/"

    return redirect(next_url)


def create_blueprint(state, import_name):
    """Create the SSO SAML extension blueprint."""
    bp = Blueprint(
        "sso_saml",
        import_name,
        url_prefix=state.url_prefix,
        template_folder="templates",
    )

    bp.add_url_rule(state.metadata_url, endpoint="metadata", view_func=metadata)

    bp.add_url_rule(
        state.sso_url, methods=["GET", "POST"], endpoint="sso", view_func=sso
    )

    bp.add_url_rule(
        state.acs_url, methods=["GET", "POST"], endpoint="acs", view_func=acs
    )

    bp.add_url_rule(
        state.slo_url, methods=["GET", "POST"], endpoint="slo", view_func=slo
    )

    bp.add_url_rule(state.sls_url, endpoint="sls", view_func=sls)

    return bp



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test utils."""

import pytest
from flask import request
from werkzeug.datastructures import MultiDict

from invenio_saml.utils import prepare_flask_request


@pytest.mark.parametrize(
    "test_request_ctx,expected",
    [
        (
            {},
            {
                "get_data": MultiDict([]),
                "http_host": "localhost",
                "https": "off",
                "post_data": MultiDict([]),
                "script_name": "/",
                "server_port": None,
            },
        ),
        (
            {"base_url": "https://tests.com:5000", "path": "/bar"},
            {
                "get_data": MultiDict([]),
                "http_host": "tests.com:5000",
                "https": "on",
                "post_data": MultiDict([]),
                "script_name": "/bar",
                "server_port": 5000,
            },
        ),
    ],
)
def test_prepare_flask_request(base_app, test_request_ctx, expected):
    """Test prepare flask request."""
    with base_app.test_request_context(**test_request_ctx):
        res = prepare_flask_request(request)
        assert res == expected



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Esteban J. G. Gabancho.
#
# invenio-saml is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio SAML errors."""


class IdentityProviderNotFound(Exception):
    """Raised when the identity provider is not found in the configuration."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2024 Esteban J. Garcia Gabancho.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

import importlib_resources as resources
import pytest
from flask import Flask
from mock import patch

from invenio_saml import InvenioSSOSAML
from invenio_saml.errors import IdentityProviderNotFound
from invenio_saml.proxies import current_sso_saml


def test_version():
    """Test version import."""
    from invenio_saml import __version__

    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioSSOSAML(app)
    assert "invenio-sso-saml" in app.extensions

    app = Flask("testapp")
    ext = InvenioSSOSAML()
    assert "invenio-sso-saml" not in app.extensions
    ext.init_app(app)
    assert "invenio-sso-saml" in app.extensions


def test_app_config(appctx):
    """Test app settings builder."""
    settings_idp1 = current_sso_saml.get_settings("test-idp")

    assert settings_idp1["strict"]
    assert settings_idp1["idp"]["entityId"] == "https://test-idp.com"

    # Handlers
    assert callable(current_sso_saml.get_handler("test-idp", "login_handler"))
    assert callable(current_sso_saml.get_handler("test-idp", "acs_handler"))
    assert current_sso_saml.get_handler("test-idp", "settings_handler") is None

    settings_idp2 = current_sso_saml.get_settings("idp-file")

    assert settings_idp2["idp"]["entityId"] == "https://login.idp.com"
    assert settings_idp2["sp"]["x509cert"] == "crt\n"
    assert settings_idp2["sp"]["privateKey"] == "key\n"

    with (resources.files(__name__) / "data" / "idp.xml").open() as f:
        response = f.read()

    with patch("onelogin.saml2.idp_metadata_parser.urllib2.urlopen") as urlopen_mock:
        urlopen_mock.return_value = type(
            "Response", (), {"read": lambda *args, **kwargs: response}
        )()
        settings_idp2 = current_sso_saml.get_settings("idp-url")
        assert settings_idp2["idp"]["entityId"] == "https://login.idp.com"


def test_auth(appctx, metadata_response):
    """Test Auth class."""
    with appctx.test_request_context(), pytest.raises(IdentityProviderNotFound):
        auth = current_sso_saml.get_auth("wrong-idp")

    with appctx.test_request_context():
        auth = current_sso_saml.get_auth("test-idp")
        assert auth
        assert auth.idp == "test-idp"
